/* This is source code for Remote Master Daemon.
   Copyright (c) 1998 by Oleg Drokin <green@ccssu.crimea.ua>
   
   This program is a free software; you can freely redistribute it.
*/

#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <syslog.h>
#include <sys/vt.h>
#include <linux/tty.h>
#include <linux/cdrom.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <signal.h>


#include "remoted.h"

int remote_keys[]={ key_ZOOM, \
	key_FREEZE, \
	key_FUNC, \
	key_EXIT, \
	key_1, \
	key_2, \
	key_3, \
	key_VOL_UP, \
	key_4, \
	key_5, \
	key_6, \
	key_VOL_DOWN, \
	key_7, \
	key_8, \
	key_9, \
	key_CH_UP, \
	key_0, \
	key_ENTER, \
	key_MUTE, \
	key_CH_DOWN, \
	key_VIDEO, \
	key_STYLE, \
	key_CHANGE, \
	key_JUMP, \
	key_REW, \
	key_PLAY, \
	key_FF, \
	key_EJECT, \
	key_STOP, \
	key_PAUSE };
char *remote_commands[NR_COMMANDS];
char *config_keys[]={ "key_ZOOM", \
	"key_FREEZE", \
	"key_FUNC", \
	"key_EXIT", \
	"key_1", \
	"key_2", \
	"key_3", \
	"key_VOL_UP", \
	"key_4", \
	"key_5", \
	"key_6", \
	"key_VOL_DOWN", \
	"key_7", \
	"key_8", \
	"key_9", \
	"key_CH_UP", \
	"key_0", \
	"key_ENTER", \
	"key_MUTE", \
	"key_CH_DOWN", \
	"key_VIDEO", \
	"key_STYLE", \
	"key_CHANGE", \
	"key_JUMP", \
	"key_REW", \
	"key_PLAY", \
	"key_FF", \
	"key_EJECT", \
	"key_STOP", \
	"key_PAUSE" };

char *device=NULL;
char *cdrom=NULL;
char **devices[]={&device,&cdrom};
char *config_names[]={"device", "cdrom"};

int cons_fd; /* File number of console */
FILE *fd; /* File of remote device */
struct termios it_save;

/* This is needed for CD-ROM playing support */
struct cdrom_ti ti; /* Needed by CD-ROM IOCTLs to play */
struct cdrom_tochdr tochdr;  /* Table of contents of CD */
struct cdrom_subchnl sc;     /* CD-ROM Drive status */
int cd_fd;

int mixer_fd, store_volume=0; /* this is vor volume changing support */

void internal_event(int ch);

void read_config(void)
{
/* Parse configuration file. */
	FILE *f;
	char buf[4000], *ptr, *ptr1;
	int i,t;

	/* clearing commands */
	for (i=0;i<NR_COMMANDS;i++) remote_commands[i]=NULL;

	f = fopen(CONFIG_FILE, "r");
	if (f == NULL) {
		printf("remoted: Configuration file %s not found.\n",
			CONFIG_FILE);
		exit(1);
	}
	fgets(buf,4000,f);

	while( !feof(f)) {

#ifdef CONF_DEBUG
		printf("parsing: %s\n",buf);
#endif
		if (buf[0] == '#' || buf[0] == 0) { /* Comment or empty line */
			fgets(buf,4000,f); 
			continue;
		}
		ptr=index(buf,' ');
		ptr1=index(buf,'\t');
		if (ptr1 == NULL && ptr == NULL) {t=strlen(buf); ptr=buf+t; }
			else if (ptr1 == NULL) t=ptr-buf;
				else if (ptr == NULL) { t=ptr1-buf; ptr=ptr1; }
					else {
						ptr=(ptr<ptr1)? ptr:ptr1;
						t=ptr-buf;
					}
		
		for(i=0;i<NR_COMMANDS;i++) {
			if (strncasecmp(buf,config_keys[i],t) == 0) {
				while ( *ptr == 0 || isspace(*ptr) ) ptr++;
				ptr1=strlen(buf)-1+buf;
				while ( *ptr1 == 0 || isspace(*ptr1) ) *ptr1--=0;
				 if (*ptr != 0 ) {
					remote_commands[i]=strdup(ptr);
#ifdef CONF_DEBUG
                			printf("parsed: %s as %s\n",config_keys[i], ptr);
#endif
					break;
				}
			}
		}
		if (i == NR_COMMANDS) { /* no one command found, then scan 
					for non-keyboard command */
			for(i=0;i<2;i++) {
				if (strncasecmp(buf,config_names[i],t) == 0) {
					while ( *ptr == 0 || isspace(*ptr) ) ptr++;
                                	ptr1=strlen(buf)-1+buf;
                                	while ( *ptr1 == 0 || isspace(*ptr1) ) *ptr1--=0;
					if (*ptr !=0 ) {
                                        	*devices[i]=strdup(ptr);
#ifdef CONF_DEBUG
                                        	printf("parsed: %s as %s\n",config_names[i], ptr);
#endif
                                        	break;
                                	}
                        	}
			}
                }

		fgets(buf,4000,f);
	}
	fclose(f);
}

void signal_handler(int sig)
{
	int i;

	if (sig==SIGHUP) {
		syslog(LOG_DAEMON | LOG_WARNING, "Reloading config file.\n");
		for (i=0;i<NR_COMMANDS;i++) if (remote_commands[i] != NULL) free(remote_commands[i]);
		free(device);free(cdrom);
		device = cdrom = NULL;
		read_config();
		if (!device) { /* no device line found in config file */
	                syslog(LOG_DAEMON | LOG_ERR, "No Device line in config file" CONFIG_FILE ", exiting...\n");
	                exit (1);
		}
		if (!cdrom) { /* no cdrom line in config file */
                	cdrom=strdup("/dev/cdrom");
                }

		fclose(fd);
		fd = fopen(device, "r");
	        if (fd == NULL) {
	                syslog(LOG_DAEMON | LOG_ERR, "cannot open %s, exiting...\n", device);
	                perror(device);
	                exit(1);
	        }
		signal(SIGHUP,signal_handler);
//		return 1;
        } else {
		syslog(LOG_DAEMON | LOG_ERR, "Killed by signal %d!",sig);
		do_exit();
	}
}
		
int main(void) 
{

	struct termios it;
	int ch,i,pid;
	struct stat buf;
	FILE *pid_fd;

	printf("RemoteD v." VERSION " started.\n");
	if (stat(PID_FILE, &buf) >= 0 ) {
		printf(PID_FILE " exists. Probably you have another RemoteD running\n");
		exit(1);
	}
	
	read_config();
	if (!device) { /* no device line found in config file */
		printf("remoted: ERROR: No Device line in config file" CONFIG_FILE "\n" );
		exit (1);
	}

	if (!cdrom) { /* no cdrom line in config file */
		cdrom=strdup("/dev/cdrom");
		}

        if ((cons_fd = open("/dev/console",O_WRONLY)) < 0) {
                perror("remoted: Can't open /dev/console\n");
                exit(1);
        }

        fd = fopen(device, "r");
        if (fd == NULL) {
                printf("remoted: cannot open %s\n", device);
		perror(device);
                exit(1);
        }

	mixer_fd = open("/dev/mixer", O_RDWR);
	
/*  Turning ON raw mode on device */
	
	tcgetattr(fileno(fd),&it_save); /* Save old settings */

	cfmakeraw(&it);
	cfsetospeed(&it, B9600);
	cfsetispeed(&it, B9600);

	tcsetattr(fileno(fd),TCSANOW,&it);
/* Raw mode is ON */

	if (geteuid() && getuid())
        {
                printf("remoted: must be run as root.\n");
                exit(1);
        }
	
	if ((pid = fork()) < 0)
        {
                perror("remoted: Couldn't fork()");
                exit(1);
        }
        else
                if (pid != 0)
                        exit(0);

        /* Here we are daemon */

        setsid();
        chdir("/");
        umask(0133);
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
	openlog("RemoteD",LOG_PID,LOG_DAEMON);

	if ((pid_fd=fopen(PID_FILE,"w")) == NULL) {
		syslog(LOG_DAEMON | LOG_ERR, "Cannot write to " PID_FILE ". exiting...");
		do_exit();
	}
	fprintf(pid_fd,"%d",getpid());
	fclose(pid_fd);

	for (i=1;i<=15;i++) 
		signal(i,&signal_handler);
	
	while(1) {
		if ((ch=fgetc(fd))==-1) continue; /* Silently ignore SIGHUP */
		for(i=0;i<NR_COMMANDS;i++) {
			if ( remote_keys[i] == ch) {
#ifdef DEBUG
				syslog(LOG_DAEMON | LOG_DEBUG, "%s pressed\n",config_keys[i]);
#endif
				if (remote_commands[i] != NULL) {
#ifdef DEBUG
					syslog(LOG_DAEMON | LOG_DEBUG, "executing %s\n",remote_commands[i]);
#endif
					system(remote_commands[i]); 
				} else 
					internal_event(ch);
				break;
			}
		}
		if (i == NR_COMMANDS ) {
			syslog(LOG_DAEMON | LOG_WARNING, "Invalid keycode received 0x%x\n",ch);
		}
	}

	return 0;
}

void do_exit(void)
{
	tcsetattr(fileno(fd), TCSANOW, &it_save);
	close(cons_fd);
	fclose(fd);
	closelog();
	unlink(PID_FILE);
	exit(0);
}

int scan_cd(void)
{
	cd_fd = open(cdrom, O_RDONLY);
	if ( cd_fd>=0 ) {
		ioctl(cd_fd,CDROMREADTOCHDR, &tochdr);
		sc.cdsc_format = CDROM_MSF;
		ioctl(cd_fd,CDROMSUBCHNL,&sc);
	} else {
		return 1;
	}
	return 0;
}

/*  play CD. */
void play_cd(int offset)
{
	if ( scan_cd() == 0 ) {
		ti.cdti_trk0 = (sc.cdsc_trk + offset < tochdr.cdth_trk0) ? \
			tochdr.cdth_trk1 : (sc.cdsc_trk + offset > \
				tochdr.cdth_trk1) ? tochdr.cdth_trk0 : sc.cdsc_trk + offset; /* starting track */
		
		ti.cdti_trk1 = tochdr.cdth_trk1; /* Last track */
		ti.cdti_ind0 = ti.cdti_ind1 = 0; /* Indexes */
		
		ioctl(cd_fd,CDROMSTOP); /* Stop motor! */
		ioctl(cd_fd,CDROMPLAYTRKIND, &ti); /* Go!!! ;) */
		close(cd_fd);
	} 
}

void eject_cd()
{
	if (scan_cd() == 0) {
		ioctl(cd_fd,CDROMSTOP);
		ioctl(cd_fd,CDROMEJECT);
		close(cd_fd);
	}
}

void stop_cd()
{
	if (scan_cd() == 0) {
                ioctl(cd_fd,CDROMSTOP);
		close(cd_fd);
	}
}

void pause_cd()
{
	if (scan_cd() == 0) {
		ioctl(cd_fd, sc.cdsc_audiostatus == CDROM_AUDIO_PAUSED ? \
			CDROMRESUME : CDROMPAUSE);
		close(cd_fd);
	}
}

void chvt(int num)
{
	if (ioctl(cons_fd, VT_ACTIVATE, num) >= 0)
        	(void) ioctl(cons_fd, VT_WAITACTIVE, num);
}

void chvt_rel(int offset)
{
	struct ioc_lin {
		int twelve;
		int nothing;
	} io_data = { 12, 0};

#ifdef VT_DEBUG
	syslog(LOG_DAEMON | LOG_DEBUG, "TIOCLINUX reported current console %d\n",ioctl(cons_fd, TIOCLINUX, &io_data));
#endif

	chvt(ioctl(cons_fd, TIOCLINUX, &io_data) + offset + 1);
}

void ch_volume(int offset)
{
	int volume;

	if (mixer_fd>=0) {
		ioctl(mixer_fd, SOUND_MIXER_READ_VOLUME, &volume);
		volume = ((volume >> 8) & 0x7f) + offset;
		volume = volume < 0 ? 0 : volume > 100 ? 100 : volume;
		volume |= volume << 8;
		ioctl(mixer_fd, SOUND_MIXER_WRITE_VOLUME, &volume);
	}
}

void mute()
{
	if (mixer_fd>=0) {
		if (store_volume==0) {
                	ioctl(mixer_fd, SOUND_MIXER_READ_VOLUME, &store_volume);
			ch_volume(-100);
		} else {
			ioctl(mixer_fd, SOUND_MIXER_WRITE_VOLUME, &store_volume);
			store_volume=0;
		}
	}
}
		
void internal_event(int ch)
{
	switch (ch) {

		case key_EXIT:
			syslog(LOG_DAEMON | LOG_WARNING, "Exit requested from Master pad. Quitting\n");
			do_exit();

		case key_1:
			chvt(1); break;

		case key_2:
			chvt(2); break;
		case key_3:
			chvt( 3); break;
		case key_4:
			chvt(4); break;
		case key_5:
			chvt(5); break;
		case key_6:
			chvt(6); break;
		case key_7:
			chvt(7); break;
		case key_8:
			chvt(8); break;
		case key_9:
			chvt(9); break;
		case key_0:
			chvt(10); break;

		case key_CH_UP: 
			chvt_rel(1);
			break;
		case key_CH_DOWN:
			chvt_rel(-1);
			break;

		case key_VOL_UP:
			ch_volume(5);
			break;

		case key_VOL_DOWN:
			ch_volume(-5);
                        break;

		case key_MUTE:
			mute();
			break;

		case key_REW:
			play_cd(-1);
			break;

		case key_PLAY:
			play_cd(0);
			break;

		case key_FF:
			play_cd(1);
			break;

		case key_EJECT:
			eject_cd();
			break;

		case key_STOP: 
			stop_cd();
			break;

		case key_PAUSE:
			pause_cd();

		case key_ZOOM:
		case key_FREEZE:
		case key_FUNC:
		case key_ENTER:
		case key_VIDEO:
		case key_STYLE:
		case key_CHANGE:
			break;
	}
return;
}


